/* === COPYRIGHT 2000 (C) ========================= TOM HENDRICK ===

   FILE:                                                      main.c

   DESCRIPTION:
      Provides the functionality of the main module of the program
      including the main initilization of the program as well as the
      functions for handling UNIX signals for SIG_INT, SIG_HUP, and
      SIG_USR1.

   $REVISION:$
   
   LICENSE:
      This file, and the files included within the tipxd project are
      distributed under the GNU Public License(GPL).  Please read the
      LICENSE file included with this distribution or consult the license
      on the web at http://www.opensource.org/licenses/gpl-license.html

   =========== T I P X D . S O U R C E F O R G E . N E T =========== */

/* -----------------------------------------------------------------
   PREPROCESSOR INCLUDES
   ----------------------------------------------------------------- */
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <syslog.h>

#include <getopt.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <signal.h>

#include "main.h"
#include "log.h"

/* -----------------------------------------------------------------
   GLOBAL OR OTHERWISE EXPORTED DATA
   ----------------------------------------------------------------- */
const char *default_config = "/etc/tipxd.conf";
sysinfo_t sysinfo;

selectinfo_t *selectinfo;

/* -----------------------------------------------------------------

   FUNCTION:                                                 main( )

   DESCRIPTION:
      Main entry point into the program from it's inception.  Arguments
      are dictacted by the system, though it is up to us to parse the
      command-line options as we see fit.

   ARGUMENTS:
      (SYSTEM)int argc     - Argument count
      (SYSTEM)char *argv[] - argument strings

   RETURN VALUE:
      int - 0 on SUCCESS, -1 on ANY ERROR

   ----------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
   char c;   /* Argument Parsing */
   int option_index = 0;
   static struct option long_options[] = {
   	{"config-file",1,NULL,'f'},
   	{"stderr",0,NULL,'e'},
   	{"verbose",0,NULL,'v'},
   	{"help",0,NULL,'h'}
   };

   /* Argument processing, set default values */
   memset( &sysinfo, '\0', sizeof( sysinfo_t ) );
   sysinfo.argc = argc;
   sysinfo.argv = argv;
   sysinfo.config_filename = default_config;
   sysinfo.opt_flags = 0;

   while ((c = getopt_long(argc,argv,"f:evh",long_options,&option_index)) != -1) {
       switch (c) {
           case 'f':
               if (!(sysinfo.config_filename = malloc(strlen(optarg)))) {
                   fprintf(stderr,"Could not allocate memory for filename storage\n");
                   exit(1);
               }
               strcpy((char *)sysinfo.config_filename,optarg);
               break;
           case 'e':
               sysinfo.opt_flags |= OPT_STDERR;
               break;
           case 'v':
               sysinfo.opt_flags |= OPT_VERBOSE;
               break;
           case 'h':
           default:
               printf("\t-f/--config-file <filename>  Use a different config file\n");
               printf("\t-e/--stderr                  Output errors to stderr instead of syslog\n");
               printf("\t-v/--verbose                 Extra information for those debugging moments.\n");
               printf("\t-h/--help                    This help.\n");
               exit(1);
   	}
   }

	openlog( "tipxd", LOG_PID | LOG_CONS, LOG_DAEMON );
   
   tipxd_log( LOG_INFO, "Started" );

   /* Okay, print out configuration information */
   tipxd_log( LOG_INFO, "Config file is %s\n", sysinfo.config_filename );

   /* Install signal handlers */
   signal( SIGUSR1, sig_usr1 );

   /* Read the configuration file */
   yyin = fopen( sysinfo.config_filename, "r+" );
   if( !yyin )
   {
      fprintf( stderr, "Unable to open configuration file : %s\n\n",
               strerror(errno) );
      exit( -1 );
   }

   yylex( );

   /*
    * Need to verify that the configuration scanned is valid, and that all
    * required components have been found.  If not, we abort with a message
    * that says why.
    */

   /*
    * Now, we need to attempt to open the logfile, and to dump the intial
    * configuration to it.
    */
   if( sysinfo.log_filename != NULL )
   {
      sysinfo.log_fp = fopen( sysinfo.log_filename, "w+" );
      /* Check for errors */
      if( NULL == sysinfo.log_fp )
      {
         /* Ooops, no log file.  Not fatal, but report it */
         tipxd_log( LOG_ERR, "Unable to open logfile %m" );
      }
      else
      {
         /* Dump the current configuration to the logfile */
         peerinfo_t *p;
         etherinfo_t *e;
         routerinfo_t *r;

         fprintf( sysinfo.log_fp, " ===== PEERS =====\n" );
         for( p = sysinfo.peerinfo; p != NULL; p = p->next )
         {
            dump_peer( sysinfo.log_fp, p );
         }
         fprintf( sysinfo.log_fp, " ===== DEVICES =====\n" );
         for( e = sysinfo.etherinfo ; e != NULL; e = e->next )
         {
            dump_ether( sysinfo.log_fp, e );
         }
         fprintf( sysinfo.log_fp, " ===== ROUTERS =====\n" );
         for( r = sysinfo.routerinfo; r != NULL; r = r->next )
         {
            dump_router( sysinfo.log_fp, r );
         }
         fprintf( sysinfo.log_fp, " ===== END CONFIGURATION =====\n\n" );

         fflush( sysinfo.log_fp );
      }
   }

   /* Create the selectinfo list */
   form_select_list( );

   /*
    * This is the main structure of the program, a nice big select() loop
    * on all the open file descriptors.  Most of the complexity should be
    * abstracted away in the individual socket operations.  Structure looks
    * like this:
    *    [1] Attempt to connect all unconnected sockets which need to be
    *    connected.
    *    [2] select( ) on the file descriptor sets.
    *    [3] For any FD in the read_set, perform the read_operation on the
    *    socket's selectinfo.
    *    [4] For any FD in the error_set, perform the error_operation on
    *    the socket's selectinfo.
    */
   for( ; ; )
   {
      selectinfo_t *s;
      int max = 0;
      fd_set read_fds, error_fds;
      struct timeval select_timeout;
      int retval;

      FD_ZERO( &read_fds );
      FD_ZERO( &error_fds );

      /* Loop through the list of file descriptor selectinfo structures and
       * find all the non-open sockets and attempt to open them */
      for( s = selectinfo; s != NULL; s = s->next )
      {
         if( s->fd == -1 )
         {
            /* Attempt to open this socket */
            s->open_op( s, s->arg );
         }

         /* This should handle the case of the open operation succeeding */
         if( s->fd > -1 )
         {
            /* Add this socket to the list for select */
            FD_SET( s->fd, &read_fds );
            FD_SET( s->fd, &error_fds );
            if( max < s->fd ) { max = s->fd; }
         }
      }

      /* Okay, now set up for the select */
      select_timeout.tv_sec = 1;
      select_timeout.tv_usec = 0;

      retval = select( max + 1, &read_fds, NULL, &error_fds, &select_timeout );

      /* Did we succeed with out select? */
      if( retval <= 0 )
      {
         continue;
      }

      /* Okay, at least one or more of the sockets was selected */
      for( s = selectinfo; s != NULL; s = s->next )
      {
         /* Is this socket even open? */
         if( s->fd > -1 )
         {
            if( FD_ISSET( s->fd, &read_fds ) )
            {
               s->read_op( s, s->arg );
            }
            if( FD_ISSET( s->fd, &error_fds ) )
            {
               s->error_op( s, s->arg, errno );
            }
         }
      }
   }

   /* Close the logfile */
   fclose( sysinfo.log_fp );

   if (sysinfo.opt_flags & OPT_STDERR)
       closelog( );

   exit( 0 );
}


/* -----------------------------------------------------------------
   
   FUNCTION:                                       verify_sysinfo( )

   DESCRIPTION:
      verifies a sysinfo_t structure is completely filled in.

   ARGUMENTS:
      N/A

   RETURN VALUE:
      int - -1 on error, 0 on valid.

   ----------------------------------------------------------------- */
int verify_sysinfo( )
{
   /* Ensure there are data and control ports */
   if( sysinfo.data_listen_port == 0 )
   {
      fprintf( stderr, "Non-zero DATA port must be specified!\n" );
      return -1;
   }

   /* Ensure there are data and control ports */
   if( sysinfo.control_listen_port == 0 )
   {
      fprintf( stderr, "Non-zero CTRL port must be specified!\n" );
      return -1;
   }

   /* Ensure there is an appropriate socket type */
   if( ( sysinfo.data_socket_type != SOCK_STREAM ) &&
       ( sysinfo.data_socket_type != SOCK_DGRAM ) )
   {
      fprintf( stderr, "No data socket type specified!\n" );
      return -1;
   }
   /* Ensure there is an appropriate socket type */
   if( ( sysinfo.control_socket_type != SOCK_STREAM ) &&
       ( sysinfo.control_socket_type != SOCK_DGRAM ) )
   {
      fprintf( stderr, "No ctrl socket type specified!\n" );
      return -1;
   }

   if( sysinfo.site_string == NULL )
   {
      fprintf( stderr, "Site String not set!\n" );
   }

   return 0;
}

/* -----------------------------------------------------------------
   
   FUNCTION:                                              sig_int( )

   DESCRIPTION:
      Signal handler for the UNIX SIGINT signal, typically delivered from
      the terminal when the user presses Control-C.  This should cleanly
      close all connections, close all ethernet devices, and shut down the
      program cleanly.

   ARGUMENTS:
      (SYSTEM)int signo - the signal number delivered to this handler,
                          which should always be equal to SIGINT.

   RETURN VALUE:
      none

   ----------------------------------------------------------------- */
void sig_int( int signo )
{
   // TODO:Need to close all open remote connections cleanly
   
   // TODO:Need to close all IPX ethernet devices cleanly

   // TODO:Log message

   exit( 0 );
}

/* -----------------------------------------------------------------

   FUNCTION:                                              sig_hup( )

   DESCRIPTION:
      Signal handler for the UNIX SIGHUP signal, typically used to
      represent a request to hang up.  In tipxd, SIGHUP will cause the
      program to shut down, and then re-read it's configuration file from
      the user's perspective.  From our perspective, it will simply shut
      down this running program, and then exec tipxd over the top of itself
      with identical arguments.

   ARGUMENTS:
      (SYSTEM)int signo - the signal number delivered to this handler,
                          which should always be equal to SIGHUP.

   RETURN VALUE:
      none

   ----------------------------------------------------------------- */
void sig_hup( int signo )
{
   // TODO:Close all open remote connections cleanly

   // TODO:Need to close all IPX ethernet devices cleanly

   // TODO:Log Message

   // Exec tipxd with the same argument set, shut down if error on exec
   execv( sysinfo.argv[0], &(sysinfo.argv[1]));

   // If we got here, there was an error
   printf( "Error on execv: %s\n", strerror(errno));
   exit( -1 );
}

/* -----------------------------------------------------------------

   FUNCTION:                                             sig_usr1( )

   DESCRIPTION:
      Signal hander for UNIX Signal SIGUSR1, reserved for use by a
      program to do something program specific.  When we receive this
      command, we are going to immediately dump the peer registration
      table, the currently open remote sockets, and the currently open
      local ethernet interfaces to a file returned by tmpnam, and then
      print a message about where the contents were dumped to.

   ARGUMENTS:
      (SYSTEM)int signo - the signal number delivered to this handler,
                          which should always be equal to SIGUSR1.

   RETURN VALUE:
      none

   ----------------------------------------------------------------- */
void sig_usr1( int signo )
{
   FILE *fp;
   int fd;
   char template[14];
   time_t t;

   // Have to do this, because mkstemp modifies the buffer
   memset( template, '\0', 14 );
   strncpy( template, "tipxd_XXXXXX", 13 );

   // Open temporary file and verify
   fd = mkstemp( template );
   if( fd < 0 )
   {
      // Ignore this error, but log it
      tipxd_log( LOG_ERR, "Unable to open temp file for dump : %m" );
      return;
   }

   // Re-open the file using FILE *
   fp = fdopen( fd, "w+" );
   if( NULL == fp )
   {
      // Ignore this error, log it, and clean up
      tipxd_log( LOG_ERR, "Unable to fdopen temp file : %m" );
      close( fd );
      return;
   }

   // Dump a file header
   time( &t );
   fprintf( fp, "TIPXD Table Dump : %s\n\n", ctime( &t ) );

   // TODO: Dump contents of peer registration table
   // TODO: Dump contents of router table
   // TODO: Dump contents of device table

   // Close the file, and then log where things went
   fclose( fp );
   tipxd_log( LOG_INFO, "Contents of Internal Tables dumped to %s on SIGUSR1",
           template );

}
