/* === COPYRIGHT 2000 (C) ========================= TOM HENDRICK ===

   FILE:                                                      peer.c

   DESCRIPTION:
      Provides all the functionality required to handle IPX peers.

   $REVISION:$
   
   LICENSE:
      This file, and the files included within the tipxd project are
      distributed under the GNU Public License(GPL).  Please read the
      LICENSE file included with this distribution or consult the license
      on the web at http://www.opensource.org/licenses/gpl-license.html

   =========== T I P X D . S O U R C E F O R G E . N E T =========== */

/* -----------------------------------------------------------------
   PREPROCESSOR INCLUDES
   ----------------------------------------------------------------- */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "peer.h"
#include "main.h"

/* -----------------------------------------------------------------
   GLOBAL OR OTHERWISE EXPORTED DATA
   ----------------------------------------------------------------- */

/* -----------------------------------------------------------------

   FUNCTION:                                          dump_peer( )

   DESCRIPTION:
      Dumps the contents of the peerinfo structure to the given File
      descriptor in a human-readable format.

   ARGUMENTS:
      FILE *fp - the file to dump to the router information to.
      peerinfo_t *info - the structure to dump

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void dump_peer( FILE *fp, peerinfo_t *info )
{
   char *loc_string;
   if( !info ) return;

   switch( info->location )
   {
      case PEER_LOC_UNREGISTERED:
         loc_string = "UNREGISTERED"; break;
      case PEER_LOC_LOCAL:
         loc_string = "LOCAL"; break;
      case PEER_LOC_REMOTE:
         loc_string = "REMOTE"; break;
      default:
         loc_string = "UNITIALIZED"; break;
   }
   // Dump the contents of info
   fprintf( fp, "PEER: %s\n", ether_address_to_str(info->address) );
   fprintf( fp, "   Location: %s\n", loc_string );
   fprintf( fp, "   Filters : %s%s%s%s\n",
                (info->filters & FILTER_802_3)?"802.3 ":"",
                (info->filters & FILTER_802_2)?"802.2 ":"",
                (info->filters & FILTER_SNAP )?"SNAP ":"",
                (info->filters & FILTER_II   )?"II":"");
   fprintf( fp, "   Broadcast Filters : %s%s%s%s\n",
                (info->filters & FILTER_802_3_BCAST)?"802.3 ":"",
                (info->filters & FILTER_802_2_BCAST)?"802.2 ":"",
                (info->filters & FILTER_SNAP_BCAST )?"SNAP ":"",
                (info->filters & FILTER_II_BCAST   )?"II":"");
   if( info->location == PEER_LOC_LOCAL )
   {
      fprintf( fp, "   Interface: %s\n", info->destination.ether->name );
   }
   else if( info->location == PEER_LOC_REMOTE )
   {
      fprintf( fp, "   Router: %s\n", info->destination.router->hostname );
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                          send_peer( )

   DESCRIPTION:
      Dumps the contents of the peerinfo structure to the given File
      descriptor in a machine-friendly format.

   ARGUMENTS:
      FILE *fp - the file to dump to the router information to.
      routerinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_peer( int fd, peerinfo_t *info )
{
   size_t len;
   peerinfo_t p;
   char *ptr = (char *)&p;
   int retval = 0;
   int bytes;

   // Copy the information to a temporary variable
   len = sizeof( peerinfo_t );
   memcpy( &p, info, len );

   // Do some byte-order operating and blanking
   p.filters = htonl( p.filters );
   p.destination.ether = NULL;
   p.next = NULL;

   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = write( fd, ptr + bytes, len - bytes );
      if( retval < 0 ) { return; }
      bytes += retval;
   } while( bytes < len );
}


/* -----------------------------------------------------------------

   FUNCTION:                                           add_peer( )

   DESCRIPTION:

   ARGUMENTS:
      routerinfo_t *info - the info structure to add to the sysinfo

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void add_peer( peerinfo_t *info )
{
   peerinfo_t *ptr;
   if( NULL == info ) return;

   // Okay, now, see if we are adding to the end of a list, or creating it
   if( sysinfo.peerinfo == NULL )
   {
      sysinfo.peerinfo = info;
   }
   else
   {
      for( ptr = sysinfo.peerinfo; ptr->next != NULL; ptr = ptr->next );
      ptr->next = info;
      info->next = NULL;
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                            find_peer( )

   DESCRIPTION:
      Attempts to find the record in the peer registration list, returns
      NULL if not found.

   ARGUMENTS:
      char * - pointer to a buffer that contains the ethernet address of
      the peer that we are searching for.

   RETURN VALUE:
      NULL, or a pointer to the peer registration record

   ----------------------------------------------------------------- */
peerinfo_t *find_peer( char *a )
{
   peerinfo_t *ptr;

   // Okay, now, see if we are adding to the end of a list, or creating it
   if( sysinfo.peerinfo == NULL )
   {
      return NULL;
   }
   else
   {
      for( ptr = sysinfo.peerinfo; ptr->next != NULL; ptr = ptr->next )
      {
         if( memcmp( ptr->address, a, ETHER_ADDR_LEN ) == 0 )
         {
            return ptr;
         }
      }
   }
   return NULL;
}

/* -----------------------------------------------------------------

   FUNCTION:                                               route ( )

   DESCRIPTION:
      This function implements the rules we use to route the packet to the
      socket or sockets to which it needs to go.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to an integer argument, representing the index of
         the protoinfo in the etherinfo array.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void route( peerinfo_t *src_peer,
            peerinfo_t *dst_peer, void *arg, char *buf, size_t len )
{
   etherinfo_t  *e;
   protoinfo_t  *p;
   routerinfo_t *r;

   /* Sanity check */
   if( NULL == src_peer ) return;

   /* This is the case that the source is local */
   if( src_peer->location == PEER_LOC_LOCAL )
   {
      /* Unregistered desitination or broadcast */
      if( ( dst_peer == NULL ) ||
          ( dst_peer->location == PEER_LOC_UNREGISTERED ) )
      {
         /* Send to all ROUTER sockets as well as all IPX sockets that are
          * not the same as the source
          */
         send_rmt_all( NULL, buf, len );
         send_ipx_all( src_peer->destination.ether, buf, len );
      }
      else
      {
         /* We just need to send to the correct destination location,
          * wherever that may be */
         if( dst_peer->location == PEER_LOC_LOCAL )
         {
            if( src_peer->destination.ether != dst_peer->destination.ether)
            {
               int index = 0;
               /* Send to this ether */
               e = dst_peer->destination.ether;
               p = e->protocols;

               p->selectinfo->write_op( p->selectinfo, &index, buf, len );
            }
         }
         else if( dst_peer->location == PEER_LOC_REMOTE )
         {
            r = dst_peer->destination.router;

            /* IF the pointer is NULL, we use the UPD socket since there is
             * no open connection, per se, to the remote host.
             */
            if( r->data_selectinfo == NULL )
            {
               selectinfo_t *s_info = sysinfo.udp_data_selectinfo;
               s_info->write_op( s_info, r, buf, len );

            }
         }
      }
   }
   /* This peer is located on a remote router.  Now, the routing rules
    * change for these cases */
   else
   {
      if( ( dst_peer == NULL ) ||
          ( dst_peer->location == PEER_LOC_UNREGISTERED ) )
      {
         /* We need to send to ALL local IPX sockets */
         send_ipx_all( NULL, buf, len );
      }
      else if( dst_peer->location == PEER_LOC_LOCAL )
      {
         int index = 0;
         /* Send to this ether */
         e = dst_peer->destination.ether;
         p = e->protocols;

         p->selectinfo->write_op( p->selectinfo, &index, buf, len );
      }
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                         send_rmt_all( )

   DESCRIPTION:
      Sends the packet buffer given to all the remote routers registered
      via either UDP or TCP as necessary.  If the pointer exep is given,
      then it sends to all the remote routers which are not a match for the
      router given.

   ARGUMENTS:
      routerinfo_t * - a pointer to the exception router.
      char * - the packet data, complete with datapkt_t structure attached
            to the front.
      size_t - the length of the buffer, in bytes, including the given
            header.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_rmt_all( routerinfo_t *exep, char *buf, size_t s )
{
   routerinfo_t *ptr;
   selectinfo_t *selectinfo;

   /* Loop through all the routers */
   for( ptr = sysinfo.routerinfo; ptr != NULL; ptr = ptr->next )
   {
      /* Send to this router if the routerinfo is different address */
      if( ptr == exep ) { continue; }

      /* Otherwise, write to this router */
      if( ptr->data_selectinfo == NULL )
      {
         selectinfo = sysinfo.udp_data_selectinfo;
         selectinfo->write_op( selectinfo, ptr, buf, s );
      }
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                         send_ipx_all( )

   DESCRIPTION:
      Sends the packet given to all the IPX sockets.  If the pointer exep
      is not NULL, then this function sends to all the IPX sockets with the
      exception of the pointer given.

   ARGUMENTS:
      etherinfo_t * - a pointer to the exception etherinfo_t.
      char * - the packet data, complete with datapkt_t structure attached
            to the front.
      size_t - the length of the buffer, in bytes, including the given
            header.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_ipx_all( etherinfo_t *exep, char *buf, size_t s )
{
   etherinfo_t  *ptr;
   protoinfo_t  *proto;
   selectinfo_t *selectinfo;
   int index = 0;

   /* Loop through all the etherinfo */
   for( ptr = sysinfo.etherinfo; ptr != NULL; ptr = ptr->next )
   {
      /* Send to this ether if the etherinfo is different address */
      if( ptr == exep ) { continue; }

      proto = ptr->protocols;
      selectinfo = proto->selectinfo;

      selectinfo->write_op( selectinfo, &index, buf, s );
   }
}
