/* === COPYRIGHT 2000 (C) ========================= TOM HENDRICK ===

   FILE:                                                    router.c

   DESCRIPTION:
      Provides the functionality of operations on routerinfo_t structures
      including printing them in human and machine-readable formats.

   $REVISION:$
   
   LICENSE:
      This file, and the files included within the tipxd project are
      distributed under the GNU Public License(GPL).  Please read the
      LICENSE file included with this distribution or consult the license
      on the web at http://www.opensource.org/licenses/gpl-license.html

   =========== T I P X D . S O U R C E F O R G E . N E T =========== */

/* -----------------------------------------------------------------
   PREPROCESSOR INCLUDES
   ----------------------------------------------------------------- */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <syslog.h>

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>

#include "router.h"
#include "main.h"
#include "log.h"

/* -----------------------------------------------------------------
   GLOBAL OR OTHERWISE EXPORTED DATA
   ----------------------------------------------------------------- */

/* -----------------------------------------------------------------

   FUNCTION:                                          dump_router( )

   DESCRIPTION:
      Dumps the contents of the routerinfo structure to the given File
      descriptor in a human-readable format.

   ARGUMENTS:
      FILE *fp - the file to dump to the router information to.
      routerinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void dump_router( FILE *fp, routerinfo_t *info )
{
   if( !info ) return;
   // Dump the contents of info
   fprintf( fp, "NAME: %s\n", info->hostname );
   fprintf( fp, "\tFLAGS  : %s%s%s\n",
            (info->flags & ROUTER_CONNECTED )?"(CONNECTED)":"",
            (info->flags & ROUTER_CONNECT)?"(CONNECT)":"",
            (info->flags & ROUTER_ACCEPT)?"(ACCEPT)":"" );
   fprintf( fp, "\tADDR   : %s\n", inet_ntoa( info->hostaddr.sin_addr ) );
   fprintf( fp, "\tSTATS  : %ld bytes in, %ld bytes out\n",
            info->bytes_rcvd, info->bytes_sent );
   if( info->data_selectinfo != NULL )
   {
      fprintf( fp, "\tDSOCK  : %sOPEN\n",
               (info->data_selectinfo->fd < 0)?"NOT ":" " );
   }

   if( info->ctrl_selectinfo != NULL )
   {
      fprintf( fp, "\tCSOCK  : %sOPEN\n",
               (info->ctrl_selectinfo->fd < 0)?"NOT ":" " );
   }

   fprintf( fp, "\tSEQUEN : %ld\n", info->sequence );

}

/* -----------------------------------------------------------------

   FUNCTION:                                          send_router( )

   DESCRIPTION:
      Dumps the contents of the routerinfo structure to the given File
      descriptor in a machine-friendly format.

   ARGUMENTS:
      FILE *fp - the file to dump to the router information to.
      routerinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_router( int fd, routerinfo_t *info )
{
   size_t len;
   routerinfo_t r;
   char *ptr = (char *)&r;
   int retval = 0;
   int bytes;

   // Copy the information to a temporary variable
   len = sizeof( routerinfo_t );
   memcpy( &r, info, len );

   // Do some byte-order operating and blanking
   r.flags      = htonl( r.flags );
   r.host_entry = NULL;
   r.bytes_rcvd = htonl( r.bytes_rcvd );
   r.bytes_sent = htonl( r.bytes_sent );

   // Do byte-ordering on the selectinfo
   r.data_selectinfo = NULL;
   r.ctrl_selectinfo = NULL;

   r.data_port  = htons( r.data_port );
   r.ctrl_port  = htons( r.ctrl_port );
   r.sequence   = htonl( r.sequence );
   r.next       = NULL;

   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = write( fd, ptr + bytes, len - bytes );
      if( retval < 0 ) { return; }
      bytes += retval;
   } while( bytes < len );
}


/* -----------------------------------------------------------------

   FUNCTION:                                           add_router( )

   DESCRIPTION:

   ARGUMENTS:
      routerinfo_t *info - the info structure to add to the sysinfo

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void add_router( routerinfo_t *info )
{
   routerinfo_t *ptr;
   if( NULL == info ) return;

   // Okay, now, see if we are adding to the end of a list, or creating it
   if( sysinfo.routerinfo == NULL )
   {
      sysinfo.routerinfo = info;
      info->next = NULL;
   }
   else
   {
      for( ptr = sysinfo.routerinfo; ptr->next != NULL; ptr = ptr->next );
      ptr->next = info;
      info->next = NULL;
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                        verify_router( )

   DESCRIPTION:
      Verifies that the given router structure is valid, and that all the
      member variables are sane values.

   ARGUMENTS:
      routerinfo_t *info - the info structure to verify

   RETURN VALUE:
      int - 0 on success, -1 on error

   ----------------------------------------------------------------- */
int verify_router( routerinfo_t *info )
{
   int mask;
   /* Is there even a structure */
   if( NULL == info )
   {
      return -1;
   }

   /* Is there a hostname set? */
   if( info->hostname[0] == '\0' )
   {
      fprintf( stderr, "Hostname for router not set!\n" );
      return -1;
   }

   /* Ensure that there is either a CONNECT or ACCEPT flag set */
   mask = ROUTER_CONNECT | ROUTER_ACCEPT;
   if( ( info->flags & mask ) == 0 )
   {
      fprintf( stderr, "Either connect or accept must be specified!\n" );
      return -1;
   }

   /* Okay, now attempt to resolve the hostname */
   info->host_entry = gethostbyname( info->hostname );
   if( info->host_entry == NULL )
   {
      /* Ooops, unable to resolve the hostname and get address */
      fprintf( stderr, "Unable to gethostbyname( ) for host %s\n",
               info->hostname );

      fprintf( stderr, "Resolver error was %s\n", hstrerror( h_errno ) );
      return -1;
   }

   /* Okay, now we have the structure, grab necessary bits */
   memcpy( &(info->hostaddr.sin_addr.s_addr), info->host_entry->h_addr,
            (info->host_entry->h_length) );

   return 0;
}

/* -----------------------------------------------------------------

   FUNCTION:                                        open_data_udp( )

   DESCRIPTION:
      Opens the UDP data socket to the remote host, and sets all the
      necessary flags.  The argument given is a port number that this
      socket needs to be bound to.

   ARGUMENTS:
      selectinfo_t * - the file descriptor.
      void * - argument. typcast to unsigned short, and is the port
               that this socket should bind to.

   RETURN VALUE:
      int - 0 on success, -1 on error

   ----------------------------------------------------------------- */
void open_data_udp( struct _selectinfo_t *info, void *arg )
{
   /* Grab the elements */
   int fd;
   int val;
   int retval;
   unsigned short port;
   struct sockaddr_in addr;

   port = sysinfo.data_listen_port;

   /* Attempt to open the socket */
   fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
   if( fd < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to open UDP DATA socket %m" );
      return;
   }

   /* Attempt to bind this socket to the port */
   addr.sin_family = AF_INET;
   addr.sin_port   = htons( port );
   addr.sin_addr.s_addr = htonl( INADDR_ANY );

   retval = bind( fd, &addr, sizeof( addr ) );
   if( retval < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to bind UDP DATA socket : %m" );
      close( fd );
      return;
   }

   /* Okay, set the REUSEADDR flag on the UDP socket */
   retval = setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));
   if( retval < 0 )
   {
      tipxd_log( LOG_WARNING, "Unable to set REUSEADDR on UDP DATA : %m" );
   }

   tipxd_log( LOG_INFO, "UDP Data Socket open on fd %d", fd );

   info->fd = fd;
}

/* -----------------------------------------------------------------

   FUNCTION:                                       write_data_udp( )

   DESCRIPTION:
      Writes the data buffer to the socket, which is a UDP socket.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to a sockaddr structure.
      char * - the data buffer to write.
      size_t - the size in bytes of the buffer.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void write_data_udp( struct _selectinfo_t *info, void *arg, char *b, size_t s )
{
   routerinfo_t *router;
   int retval;
   size_t bytes;

   router = (routerinfo_t *)arg;

   printf( "Writing %d bytes to UDP DATA socket %s\n",
            s, router->hostname );

   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = sendto( info->fd, b + bytes, s - bytes, 0,
         (struct sockaddr *)&(router->hostaddr), sizeof( struct sockaddr ) );

      if( retval < 0 ) { info->error_op( info, arg, errno ); return; }
      bytes += retval;
   } while( bytes < s );
}

/* -----------------------------------------------------------------

   FUNCTION:                                        open_ctrl_udp( )

   DESCRIPTION:
      Opens the UDP ctrl socket to the remote host, and sets all the
      necessary flags.  The argument given is a port number that this
      socket needs to be bound to.

   ARGUMENTS:
      selectinfo_t * - the file descriptor.
      void * - argument. typcast to unsigned short, and is the port
               that this socket should bind to.

   RETURN VALUE:
      int - 0 on success, -1 on error

   ----------------------------------------------------------------- */
void open_ctrl_udp( struct _selectinfo_t *info, void *arg )
{
   /* Grab the elements */
   int fd;
   int val;
   int retval;
   unsigned short port;
   struct sockaddr_in addr;

   port  = sysinfo.control_listen_port;

   /* Attempt to open the socket */
   fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
   if( fd < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to open UDP CTRL socket %m" );
      return;
   }

   /* Attempt to bind this socket to the port */
   addr.sin_family = AF_INET;
   addr.sin_port   = htons( port );
   addr.sin_addr.s_addr = htonl( INADDR_ANY );

   retval = bind( fd, &addr, sizeof( addr ) );
   if( retval < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to bind UDP CTRL socket : %m" );
      close( fd );
      return;
   }

   /* Okay, set the REUSEADDR flag on the UDP socket */
   retval = setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));
   if( retval < 0 )
   {
      tipxd_log( LOG_WARNING, "Unable to set REUSEADDR on UDP CTRL : %m" );
   }

   tipxd_log( LOG_INFO, "UDP Control Socket open on fd %d", fd );

   info->fd = fd;
}


/* -----------------------------------------------------------------

   FUNCTION:                                       write_ctrl_udp( )

   DESCRIPTION:
      Writes the ctrl buffer to the socket, which is a UDP socket.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to a sockaddr structure.
      char * - the ctrl buffer to write.
      size_t - the size in bytes of the buffer.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void write_ctrl_udp( struct _selectinfo_t *info, void *arg, char *b, size_t s )
{
   routerinfo_t *router;
   datapkt_t *header;
   int retval;
   size_t bytes;

   router = (routerinfo_t *)arg;
   header = (datapkt_t *)b;

   header->sequence = router->sequence;
   router->sequence++;
   header->payload_size = s - sizeof( datapkt_t );

   printf( "Writing %d bytes to UDP CTRL socket %s\n",
            s, router->hostname );

   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = sendto( info->fd, b + bytes, s - bytes, 0,
         (struct sockaddr *)&(router->hostaddr), sizeof( struct sockaddr ) );

      if( retval < 0 ) { info->error_op( info, arg, errno ); return; }
      bytes += retval;
   } while( bytes < s );
}

/* -----------------------------------------------------------------

   FUNCTION:                                        read_data_udp( )

   DESCRIPTION:
      Reads a packet from the UDP socket interface given, and attempts
      to route it using the route function.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to a etherinfo_t * structure

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void read_data_udp( struct _selectinfo_t *info, void *arg )
{
   char *buf;
   size_t len;
   routerinfo_t *router;
   struct ethhdr *header;
   peerinfo_t *src_peer, *dst_peer;
   struct sockaddr_in addr;
   size_t addr_len;
   

   /* Allocate a large buffer and clear it */
   len = ETH_FRAME_LEN + sizeof( datapkt_t );
   buf = (char *)malloc( len );
   memset( buf, '\0', len );

   /* Attempt to read an entire buffer */
   len = recvfrom( info->fd, buf, len, 0, &addr, &addr_len );

   tipxd_log( LOG_INFO, "Read %d bytes from UDP: %s",
               len, inet_ntoa(addr.sin_addr) );

   /* Find the router, or drop the packet */
   for(router = sysinfo.routerinfo; router != NULL; router = router->next)
   {
      /* Compare */
      if(memcmp(&router->hostaddr.sin_addr.s_addr,&addr.sin_addr.s_addr,4)==0)
            break;
   }

   if( router == NULL )
   {
      /* Router not found, drop */
      free( buf );
      return;
   }

   header = (struct ethhdr *)&buf[sizeof(datapkt_t)];
   src_peer = find_peer( header->h_source );
   dst_peer = find_peer( header->h_dest );

   /* If the src_peer is not registered, then we nede to register it */
   if( src_peer == NULL )
   {
      src_peer = (peerinfo_t *)malloc( sizeof( peerinfo_t ) );
      memset( src_peer, '\0', sizeof( peerinfo_t ) );

      memcpy( src_peer->address, header->h_source, ETHER_ADDR_LEN );
   }

   if( src_peer->location == PEER_LOC_UNREGISTERED )
   {
      src_peer->location = PEER_LOC_REMOTE;
      src_peer->destination.router = router;
   }

   /* Attempt to route the packet */
   route( src_peer, dst_peer, NULL, buf, len);

   /* De-allocate the buffer */
   free( buf );
}
