/* === COPYRIGHT 2000 (C) ========================= TOM HENDRICK ===

   FILE:                                                    select.c

   DESCRIPTION:
      This contains all of the common operations for the selectinfo
      structures.

   $REVISION:$
   
   LICENSE:
      This file, and the files included within the tipxd project are
      distributed under the GNU Public License(GPL).  Please read the
      LICENSE file included with this distribution or consult the license
      on the web at http://www.opensource.org/licenses/gpl-license.html

   =========== T I P X D . S O U R C E F O R G E . N E T =========== */

/* -----------------------------------------------------------------
   PREPROCESSOR INCLUDES
   ----------------------------------------------------------------- */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <syslog.h>

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "main.h"
#include "select.h"
#include "router.h"
#include "log.h"

/* -----------------------------------------------------------------
   GLOBAL OR OTHERWISE EXPORTED DATA
   ----------------------------------------------------------------- */


/* -----------------------------------------------------------------

   FUNCTION:                                        close_general( )

   DESCRIPTION:
      Closes the socket associated with the given selectinfo structure.
      This function is #define'd to other names for future expansion.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to an integer argument, representing the index of
         the protoinfo in the etherinfo array.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void close_general( struct _selectinfo_t *info, void *arg )
{
   if( info->type == SELECT_TYPE_IPX )
   {
      etherinfo_t *ether;
      protoinfo_t *proto;
      int index;

      index = *((int *)arg);
      ether = (etherinfo_t *)info->data;
      proto = (protoinfo_t *)(&ether->protocols[index]);

      tipxd_log( LOG_ERR, "Closing IPX(%s) socket of %s",
            proto->name, ether->name );
   }
   else if( info->data != NULL )
   {
      routerinfo_t *router;

      router = (routerinfo_t *)info->data;

      tipxd_log( LOG_ERR, "Closing socket to %s",
            router->hostname );
   }

   /* Close the IPX socket */
   close( info->fd );

   info->fd = -1;
}

/* -----------------------------------------------------------------

   FUNCTION:                                        error_general( )

   DESCRIPTION:
      This is a general error function shared by all socket types.  This is
      #define'd to other names for future expansion.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to an integer argument, representing the index of
         the protoinfo in the etherinfo array.
      int - the current value of the errno variable.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void error_general( struct _selectinfo_t *info, void *arg, int err )
{
   /* Okay, we need to grab the informtion structures */

   if( info->type == SELECT_TYPE_IPX )
   {
      etherinfo_t *ether;
      protoinfo_t *proto;
      int index;

      index = *((int *)arg);
      ether = (etherinfo_t *)info->data;
      proto = (protoinfo_t *)(&ether->protocols[index]);

      tipxd_log( LOG_ERR, "Error detected on %s socket of %s : %s",
            proto->name, ether->name, strerror( err ) );
   }
   else if( info->data != NULL )
   {
      routerinfo_t *router;

      router = (routerinfo_t *)info->data;

      tipxd_log( LOG_ERR, "Error detected on TCP socket to %s : %s",
            router->hostname, strerror( err ) );
   }

   /* Now, switch on the value of errno */
   switch( err )
   {
      case EIO:
      case EISDIR:
      case EBADF:
      case EINVAL:
      case EFAULT:
      case EPIPE:
         {
            /* Fatal error on this socket, close and move on */
            info->close_op( info, arg );
            info->fd = -1;
         }
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                     form_select_list( )

   DESCRIPTION:
      Forms the list of selectinfo structures, and places them into the
      system selectinfo list.  This does not attempt to open the sockets,
      it just creates the structures and intializes them.

   ARGUMENTS:
      N/A

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void form_select_list(  )
{
   selectinfo_t *ptr1;
   selectinfo_t *ptr2;
   etherinfo_t  *ether;
   size_t s = sizeof( selectinfo_t );
   int i;
   int *i_ptr;
   int count = 0;

   /* Create the UDP listening socket */
   ptr1 = (selectinfo_t *)malloc( s );
   memset( ptr1, '\0', s );

   /* Initialize members */
   ptr1->fd = -1;
   ptr1->type = SELECT_TYPE_DATA_UDP;
   ptr1->read_op  = read_data_udp;
   ptr1->write_op = write_data_udp;
   ptr1->error_op = error_data_udp;
   ptr1->open_op   = open_data_udp;
   ptr1->close_op = close_data_udp;
   ptr1->data = NULL;
   ptr1->arg  = NULL;
   ptr1->next = NULL;

   /* This is the start of the list */
   sysinfo.udp_data_selectinfo = ptr1;
   selectinfo = ptr1;
   ptr2 = ptr1;

   count++;

   /* Now form the IPX selectinfo elements */
   for(ether = sysinfo.etherinfo; ether != NULL; ether = ether->next)
   {
      /* Create selectinfo for each protoinfo */
      for(i = 0; i < ether->protocount; i++)
      {
         ptr1 = (selectinfo_t *)malloc( s );
         memset( ptr1, '\0', s );

         ptr1->fd = -1;
         ptr1->type = SELECT_TYPE_IPX;
         ptr1->read_op  = read_ipx;
         ptr1->write_op = write_ipx;
         ptr1->error_op = error_ipx;
         ptr1->open_op  = open_ipx;
         ptr1->close_op = close_ipx;
         ptr1->data = ether;
         i_ptr      = malloc( sizeof( int ) );
         *i_ptr     = i;
         ptr1->arg  = i_ptr;
         ptr1->next = NULL;

         ether->protocols[i].selectinfo = ptr1;

         /* Add this to the end of the list */
         ptr2->next = ptr1;
         ptr2 = ptr1;

         count++;
      }
   }

   tipxd_log( LOG_INFO, "Added %d SelectInfo structures\n", count );
}
